/*! For license information please see toggleTheme.acf125721c09f1a04027.js.LICENSE.txt */
!function(e,t){"object"==typeof exports&&"object"==typeof module?module.exports=t():"function"==typeof define&&define.amd?define([],t):"object"==typeof exports?exports.Airflow=t():(e.Airflow=e.Airflow||{},e.Airflow.toggleTheme=t())}(self,(()=>(()=>{const e="darkTheme",t=document.documentElement,o=e=>JSON.parse(localStorage.getItem(e)),n=o=>{localStorage.setItem(e,o),t.setAttribute("data-color-scheme",o?"dark":"light")};return document.addEventListener("DOMContentLoaded",(()=>{document.getElementById("themeToggleButton").addEventListener("click",(()=>{(()=>{const t=o(e);n(!t)})()}))})),(()=>{const t=o(e);n(t)})(),{}})()));