# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from marshmallow_sqlalchemy import SQLAlchemySchema, auto_field

from airflow.models import SlaMiss


class SlaMissSchema(SQLAlchemySchema):
    """Sla Miss Schema."""

    class Meta:
        """Meta."""

        model = SlaMiss

    task_id = auto_field(dump_only=True)
    dag_id = auto_field(dump_only=True)
    execution_date = auto_field(dump_only=True)
    email_sent = auto_field(dump_only=True)
    timestamp = auto_field(dump_only=True)
    description = auto_field(dump_only=True)
    notification_sent = auto_field(dump_only=True)
