#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Adding max_consecutive_failed_dag_runs column to dag_model table.

Revision ID: 8e1c784a4fc7
Revises: ab34f260b71c
Create Date: 2024-01-18 15:02:24.587206

"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8e1c784a4fc7"
down_revision = "ab34f260b71c"
branch_labels = None
depends_on = None
airflow_version = "2.9.0"


def upgrade():
    """Apply Adding max_consecutive_failed_dag_runs column to dag_model table."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("dag", schema=None) as batch_op:
        batch_op.add_column(sa.Column("max_consecutive_failed_dag_runs", sa.Integer()))

    # ### end Alembic commands ###


def downgrade():
    """Unapply Adding max_consecutive_failed_dag_runs column to dag_model table."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("dag", schema=None) as batch_op:
        batch_op.drop_column("max_consecutive_failed_dag_runs")

    # ### end Alembic commands ###
