/*!
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

const jsonParse = (content: any) => {
  let contentJson = null;
  let contentFormatted = "";
  let isJson = false;
  try {
    if (typeof content === "string") {
      contentJson = JSON.parse(content);
    } else {
      contentJson = content;
    }
    contentFormatted = JSON.stringify(contentJson, null, 4);
    isJson = contentJson != null && typeof contentJson === "object"; // ensure numbers/bool are not treated as JSON
  } catch (e) {
    // skip
  }
  return [isJson, contentJson, contentFormatted];
};

export default jsonParse;
