# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from marshmallow import Schema, fields

from airflow.api_connexion.schemas.enum_schemas import DagStateField


class DagStatsStateSchema(Schema):
    """DagStatsState Schema."""

    state = DagStateField(dump_only=True)
    count = fields.Int(dump_only=True)


class DagStatsSchema(Schema):
    """DagStats Schema."""

    dag_id = fields.String(required=True)
    stats = fields.List(fields.Nested(DagStatsStateSchema))


class DagStatsCollectionSchema(Schema):
    """DagStassCollection Schema."""

    dags = fields.List(fields.Nested(DagStatsSchema))
    total_entries = fields.Int()


dag_stats_state_schema = DagStatsStateSchema()
dag_stats_schema = DagStatsSchema()
dag_stats_collection_schema = DagStatsCollectionSchema()
